//
// Created by lun on 4/13/21.
//

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
//#include <unistd.h>
#include <time.h>
//#include <sys/sysinfo.h>

#include "virus.h"
#include "haplotype.h"
#include "edges.h"

//#define DEBUG
#define DEBUG_TIME

#define LENGTH_OF(x) (sizeof(x)/sizeof(*(x)))

extern int* pStep;

void help();
void readNumOfVirusesFromMut(const char * fin, int * numOfViruses);
void readMutaions(const char * fin, bool * mutKeep, const int maxLengthOfSeq, int numOfViruses, struct Virus * viruses);
void showVirus(struct Virus * viruses, int numOfViruses);
void virus_init(struct Virus * viruses, int numOfviruses);
void hap_show(struct Hap * pHaps, int numOfHaps);
void edges_show(struct Edge * edges, int numOfEdges);
char * creat_network(struct Virus * viruses, int numOfViruses, struct Hap * haps, int numOfHaps, struct Edge * edges, int numOfEdges, int * network, int numOfEdgesInNet, int heightOfTree, bool clustering);
void countryDistribution(struct Hap * haps, int numOfHaps, struct Virus * viruses, int numOfViruses);
//void output(const char * str_folder, char * json, long memoryUsed, double time);
void free_viruses(struct Virus ** viruses, int numOfViruses);
void readMeta_hash(const char * fin, int numOfViruses, struct Virus * viruses);
int cmp_time_virus(const void* _a, const void* _b);
void treeHeight(struct Hap * haps, int numOfHaps, int * height);
void show_mutation(struct Mutation *pMut, int numOfIndel);

bool checkUsageAll(const char * str_vcf, const char * str_mut, const char * str_sta_anno, const char * str_meta, const char * str_minFreq, const char * str_maxvirus, const char * str_folder,  double * minFreq, int * maxvirus );
#ifdef MUTILTHREADING
void calcDistDirected_sorthap(struct Hap * haps, int numOfHaps, struct Edge * edges, int * numOfEdges, int nThread);
#else
void calcDistDirected_sorthap(struct Hap * haps, int numOfHaps, struct Edge * edges, int * numOfEdges );
#endif
void calcDisAll(int *network, int numOfEdgesInNet, struct Edge * edges, int *disAll);
void calcSrcSizeAve(int *network, int numOfEdgesInNet, struct Edge *edges, double *aveOfSourceSize);
void calcNumOfEdgesTimeGood(int *network, int numOfEdgesInNet, struct Edge *edges, int *numOfEdgesTimeGood);
void findHap_hashmap(struct Virus * viruses, int numOfViruses, struct Hap ** haps, int * numOfHaps);
void free_haps2(struct Hap ** pHaps, int numOfHaps);
typedef int(*Callback)(const char*, size_t, size_t, size_t, void*);
int vcf2mut(const char* vcfName, const char* mutName, Callback lineHandler, void* arg, const char* missingValue);
void readVCF(const char* str_vcf, const char* str_mut, bool * mutKeep, const int maxLengthOfSeq, int* numOfViruses, struct Virus ** viruses);
int main_siteMask(int argc, char* argv[]);
int main_fastCommunity(int argc, char* argv[]);
int getLine(char **lineptr, size_t *n, FILE *stream, char delim);
char* getFullPathName(const char* str_folder, const char* filename);
void output_json(const char * str_folder, char * filename, char * json);
int outputHapsTSV(const char* str_folder, const char* filename, struct Virus* viruses, int numOfViruses, struct Hap* haps, int numOfHaps, struct Edge* edges, int* network, int numOfEdgesInNet, bool clustering);
int outputEdgesTSV(const char* str_folder, const char* filename, struct Virus* viruses, int numOfViruses, struct Hap* haps, int numOfHaps, struct Edge* edges, int* network, int numOfEdgesInNet, bool clustering);
int outputGraphML(const char* str_folder, const char* filename, struct Virus* viruses, int numOfViruses, struct Hap* haps, int numOfHaps, struct Edge* edges, int* network, int numOfEdgesInNet, bool clustering);

#ifdef MUTILTHREADING
void hapNetDirected(const char * str_vcf, const char * str_mut, const char * str_meta, const char * str_minFreq, const char * str_maxvirus, const char * str_folder, char *str_sitemask, int* outFrmt, bool clustering, int nThread) {
#else
void hapNetDirected(const char * str_vcf, const char * str_mut, const char * str_meta, const char * str_minFreq, const char * str_maxvirus, const char * str_folder, char *str_sitemask, int* outFrmt, bool clustering) {
#endif
    // const char * str_mut, input file mutations
    // const char * str_sta_anno, input file states [optional]
    // const char * str_meta, input file metadata
    // const char * str_minFreq, filtered out SNPs with frequency < minFreq [optional]
    // const char * str_maxvirus, maximum number of virus wanted to read [optional]
    // const char * str_folder, output folder

    //for memory
    //struct sysinfo s_info;
    //long freeram_max;
    //long freeram_min;
    //long freeswap_max;
    //long freeswap_min;
    //long memoryUsed;

    //if(sysinfo(&s_info)==0){
    //    freeram_max = s_info.freeram;
    //    freeswap_max = s_info.freeswap;
    //}

    clock_t start_all, end_all;
    double duration_all;
    start_all = clock();

#ifdef DEBUG_TIME
    clock_t start, end;
    double duration;
#endif

    //some const here
    const int maxLengthOfSeq = 30000;

    //check usage and read numbers
    double minFreq = 0;
    int maxvirus = 0;
    if(false == checkUsageAll(str_vcf, str_mut, NULL, str_meta, str_minFreq, str_maxvirus, str_folder, &minFreq, &maxvirus ) ){
        //help();
        exit(EXIT_FAILURE);
    }

    char* str_mut_out = NULL;
    if (outFrmt[3])
        str_mut_out = getFullPathName(str_folder, "mut");

    if (str_vcf && outFrmt[0] + outFrmt[1] + outFrmt[2] == 0) {
        if (-1 == vcf2mut(str_vcf, str_mut_out, NULL, NULL, "*"))
            exit(EXIT_FAILURE);
        return;
    }

    int numOfViruses = 0;

    if (str_vcf) {
        numOfViruses = INT_MAX;
    }

    if (str_mut) {
        //read numOfViruses
        readNumOfVirusesFromMut(str_mut, &numOfViruses);
    }

    if (!str_sitemask) {
#ifdef DEBUG_TIME
        start = clock();
#endif
        str_sitemask = getFullPathName(str_folder, "site_mask");
        const char* argv[] = { "siteMask", ((str_vcf) ? "-f" : "-u"), ((str_vcf) ? str_vcf : str_mut), "-m", str_meta, "-q", ((str_minFreq) ? str_minFreq : "0"), "-o", str_sitemask, NULL };
        main_siteMask(LENGTH_OF(argv) - 1, (char**)argv);
#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for generate sitemask file\n", duration);
        fflush(stdout);
#endif
    }

    //mutation filter
    //mutKeep[i]: 0 (false), ignore pos i+1; 1 (true), keep pos i+1
    bool * mutKeep = (bool *)malloc(sizeof(int)*maxLengthOfSeq);
    if(mutKeep == NULL){
        fprintf(stderr, "Memory allocation failed for mutKeep\n");
        return;
    }
#ifdef DEBUG_TIME
    start = clock();
#endif
    //read site kept
    {
        for (int i = 0; i < maxLengthOfSeq; ++i) {
            mutKeep[i] = false;
        }
        FILE *fpin = fopen(str_sitemask, "r");
        if(NULL == fpin){
            fprintf(stderr, "\nError: failed to open file: %s\n", str_sitemask);
            exit(EXIT_FAILURE);
        }
        int num=1;
        while(num > 0){
            int tmp;
            num = fscanf(fpin, "%d\n", &tmp);
            if(num > 0){
                if(tmp <= maxLengthOfSeq && tmp >= 1){
                    mutKeep[tmp-1] = true;
                }
            }
        }
        fclose(fpin);
        fpin = NULL;
    }

    //mutationFilter(freqEachPos, maxLengthOfSeq, minFreq, numOfViruses, mutKeep);
#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for constructing mutation filter\n", duration);
#endif
    //reduce numOfViruses for test
    if(NULL != str_maxvirus && maxvirus > 0 && maxvirus < numOfViruses){
        numOfViruses = maxvirus;
    }
#ifdef DEBUG
    printf("numOfViruses = %d\n", numOfViruses);
#endif

    struct Virus * viruses = NULL;

    if (str_vcf) {
#ifdef DEBUG_TIME
        start = clock();
#endif
        readVCF(str_vcf, str_mut_out, mutKeep, maxLengthOfSeq, &numOfViruses, &viruses);
#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for reading mutations from vcf\n", duration);
#endif
    }

    if (str_mut) {
        //read mutations
        //only keep Snps and consider mutKeep
        viruses = (struct Virus *)malloc(sizeof(struct Virus)*numOfViruses);
        if (viruses == NULL) {
            fprintf(stderr, "Memory allocation failed for viruses\n");
            return;
        }
        virus_init(viruses, numOfViruses);
#ifdef DEBUG_TIME
        start = clock();
#endif
        readMutaions(str_mut, mutKeep, maxLengthOfSeq, numOfViruses, viruses);
#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for reading mutations\n", duration);
#endif
    }

    //read metadata
#ifdef DEBUG_TIME
    start = clock();
#endif
    readMeta_hash(str_meta, numOfViruses, viruses);
    //readMeta_3col_hash(str_meta, numOfViruses, viruses);
    //readMetaLine(str_meta, numOfViruses, viruses);
#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for reading metadata and matching with mutations\n", duration);
#endif

#ifdef DEBUG
    showVirus(viruses, numOfViruses);
    //getchar();
#endif

    //sort time, needed for calc time in each haplotype
#ifdef DEBUG_TIME
    start = clock();
#endif
    qsort(viruses, numOfViruses, sizeof(struct Virus), cmp_time_virus);
#ifdef DEBUG_TIME
    end = clock();

    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for sort all viruses (time)\n", duration);
#endif


#ifdef DEBUG
    for (int i = 0; i < numOfViruses; ++i) {
        printf("%04d %02d %02d\n", viruses[i].date->year, viruses[i].date->month,viruses[i].date->day);
    }
#endif
    int numOfVirusesALL = numOfViruses;

    //find haplotype
    struct Hap * haps = NULL;
    int numOfHaps = -1;
#ifdef DEBUG_TIME
    start = clock();
#endif
    findHap_hashmap(viruses, numOfViruses, &haps, &numOfHaps);

#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for finding haplotypes\n", duration);
    printf("virus = %d, hap = %d\n", numOfViruses, numOfHaps);
#endif
#ifdef DEBUG
    hap_show(haps, numOfHaps);
#endif

    //calculate distance (find ancestor)
    int numOfEdges = numOfHaps;
    struct Edge * edges = (struct Edge *)malloc(sizeof(struct Edge) * numOfEdges);
    if(edges == NULL){
        fprintf(stderr, "Memory allocation failed for edges\n");
        return;
    }
#ifdef DEBUG_TIME
    start = clock();
#endif
    //calcDistDirected(haps, numOfHaps, edges, &numOfEdges );
#ifdef MUTILTHREADING
    calcDistDirected_sorthap(haps, numOfHaps, edges, &numOfEdges, nThread);
#else
    calcDistDirected_sorthap(haps, numOfHaps, edges, &numOfEdges );
#endif
    //calcDistDirectedNew(haps, numOfHaps, edges, numOfEdges);
#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for calculating distances\n", duration);
#endif

#ifdef DEBUG
    edges_show(edges, numOfEdges);
    //getchar();
#endif

//construct hap net
#ifdef DEBUG_TIME
    start = clock();
#endif
    int numOfEdgesInNet = numOfHaps - 1;
    int * network = (int *)malloc(sizeof(int) * numOfEdgesInNet);
    if(network == NULL){
        fprintf(stderr, "Memory allocation failed for network\n");
        return;
    }
    int count = 0;
    for (int i = 0; i < numOfEdges; ++i) {
        if( 0 != edges[i].pHap2->numOfMut ){
            network[count] = i; //index of useful edges
            count ++;
        }
    }
    if(count != numOfEdges-1){
        fprintf(stderr, "Error: count = %d, numOfHaps = %d", count, numOfHaps);
        exit(EXIT_FAILURE);
    }
#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for constructing hap net\n", duration);
#endif

    //calculate country percentage for each haplotype
#ifdef DEBUG_TIME
    start = clock();
#endif
    countryDistribution(haps, numOfHaps, viruses, numOfViruses);
#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for calculating country percentage for each haplotype\n", duration);
#endif

    //analysis
#ifdef DEBUG_TIME
    start = clock();
#endif
    int heightOfTree;
    treeHeight(haps, numOfHaps, &heightOfTree);
    printf("heightOfTree = %d\n", heightOfTree);
    //calculate cost of the network (sum of distance, sum of number of samples in source of each edge, number of good edge according to time criterion)


    int disAll;
    calcDisAll(network, numOfEdgesInNet, edges, &disAll);
    printf("sumOfDist = %d\n", disAll);

    double aveOfSourceSize;
    calcSrcSizeAve(network, numOfEdgesInNet, edges, &aveOfSourceSize);
    printf("aveSourceSize = %f\n", aveOfSourceSize);

    int numOfEdgesTimeGood;
    calcNumOfEdgesTimeGood(network, numOfEdgesInNet, edges, &numOfEdgesTimeGood);
    printf("numOfEdgesTimeGood = %d\n", numOfEdgesTimeGood);

#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for test\n", duration);
    fflush(stdout);
#endif

    // clustering
    if (clustering) {
#ifdef DEBUG_TIME
        start = clock();
#endif
        char *str_pairs = getFullPathName(str_folder, "source_target.pairs");
        FILE *fpout = fopen(str_pairs, "w");
        if (NULL == fpout) {
            fprintf(stderr, "open file error!\n");
            exit(EXIT_FAILURE);
        }
        for (int i = 0; i < numOfEdgesInNet; ++i) {
            fprintf(fpout, "%d\t", edges[network[i]].source);
            fprintf(fpout, "%d\n", edges[network[i]].target);
        }
        fclose(fpout);
        fpout = NULL;

        char* argv1[] = {"fastCommunity", "-f", str_pairs, "-l", "firstRun", NULL};
        main_fastCommunity(LENGTH_OF(argv1) - 1, argv1);

        char stepStr[sizeof(int) * 5 / 2];
        int n = sprintf(stepStr, "%d", *pStep);
        stepStr[n] = 0;

        char* argv2[] = { "fastCommunity", "-f", str_pairs, "-l", "secondRun", "-c", stepStr, NULL };
        main_fastCommunity(LENGTH_OF(argv2) - 1, argv2);

        char *str_groups = getFullPathName(str_folder, "source_target-fc_secondRun.groups");
        FILE *fpin = fopen(str_groups, "r");
        if (NULL == fpin) {
            fprintf(stderr, "open file error!\n");
            exit(EXIT_FAILURE);
        }

        char *line = NULL;
        size_t len = 0;
        int size;
        int hap;
        int group = 0;

        while (getLine(&line, &len, fpin, '\n') != -1) {
            if (sscanf(line, "GROUP[ %*d ][ %d ]%*[^\n]\n", &size) == 1) {
                for (int i = 0; i < size; i++)
                    if (fscanf(fpin, "%d%*[^\n]\n", &hap) == 1 && hap >= 0 && hap < numOfHaps)
                        haps[hap].group = group;
                group++;
            }
        }

        fclose(fpin);
        fpin = NULL;
        free(line);
        line = NULL;
        free(str_groups);
        str_groups = NULL;
        free(str_pairs);
        str_pairs = NULL;

#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for community structure inference\n", duration);
        fflush(stdout);
#endif
    }

    //creat json
    char * json = NULL;
    if (outFrmt[0]) {
#ifdef DEBUG_TIME
        start = clock();
#endif
        json = creat_network(viruses, numOfViruses, haps, numOfHaps, edges, numOfEdges, network, numOfEdgesInNet, heightOfTree, clustering);

#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for creating json\n", duration);
#endif
#ifdef DEBUG
        printf("###network###:\n%s\n", json);
#endif
    }

    //for memory
    //if(sysinfo(&s_info)==0){
    //    freeram_min = s_info.freeram;
    //    freeswap_min = s_info.freeswap;
    //}
    //memoryUsed = freeram_max +  freeswap_max - freeram_min -  freeswap_min;// Byte

//#ifdef DEBUG
//    system("sh /home/lun/work/experiments/test-mem/mem-my.sh HapNet_v0_2_5");
//#endif

#ifdef DEBUG_TIME
    start = clock();
#endif

    //output json
    if (outFrmt[0]) {
#ifdef DEBUG_TIME
        start = clock();
#endif
        //output(str_folder, json, memoryUsed, duration_all);
        output_json(str_folder, "haplotype_loci.json", json);
        free(json);
        json = NULL;
#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for writing results in JSON format\n", duration);
#endif
    }

    //output tsv
    if (outFrmt[1]) {
#ifdef DEBUG_TIME
        start = clock();
#endif

        if (outputHapsTSV(str_folder, "haps_out.tsv", viruses, numOfViruses, haps, numOfHaps, edges, network, numOfEdgesInNet, clustering))
            exit(EXIT_FAILURE);

        if (outputEdgesTSV(str_folder, "edges_out.tsv", viruses, numOfViruses, haps, numOfHaps, edges, network, numOfEdgesInNet, clustering))
            exit(EXIT_FAILURE);

#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for writing results in TSV format\n", duration);
#endif
    }

    //output GraphML
    if (outFrmt[2]) {
#ifdef DEBUG_TIME
        start = clock();
#endif

        if (outputGraphML(str_folder, "haplotype_loci.graphml", viruses, numOfViruses, haps, numOfHaps, edges, network, numOfEdgesInNet, clustering))
            exit(EXIT_FAILURE);

#ifdef DEBUG_TIME
        end = clock();
        duration = (double)(end - start) / CLOCKS_PER_SEC;
        printf("%f s for writing results in GraphML format\n", duration);
#endif
    }

#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for writing to output file\n", duration);
#endif

#ifdef DEBUG_TIME
    start = clock();
#endif
    //free, need to free hap.*, ...,

    free(mutKeep);
    mutKeep = NULL;
    free(edges);
    edges = NULL;
    free(str_mut_out);
    str_mut_out = NULL;


    free_haps2(&haps, numOfHaps);


    free_viruses(&viruses, numOfVirusesALL);
    //free(json);
    //free(json);//!!!!!!!!!!!!!malloc_consolidate(): invalid chunk size when numOfSample is smaller than 100
    //json = NULL;
#ifdef DEBUG_TIME
    end = clock();
    duration = (double)(end - start) / CLOCKS_PER_SEC;
    printf("%f s for free\n", duration);
#endif

    //all time
    end_all = clock();
    duration_all = -((double)(start_all - end_all)) / CLOCKS_PER_SEC;
    printf("\nTotal time used: %f s\n", duration_all);
}
