//---------------------------------------------------------------------------
// Copyright (C) 2021 Bo Xu <xubo123@big.ac.cn>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Command line options parser
//---------------------------------------------------------------------------

#ifndef OPTIONPARSER_H
#define OPTIONPARSER_H

#define LENGTH_OF(x) (sizeof(x)/sizeof(*(x)))

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
    const char* name;
    int hasArg;
    int id;
} Option;

typedef struct {
    int id;
    const char* val;
} Arg;

enum {
    ESuccess = 0,
    ESuccessEnd,
    ERepeated,
    EInvalid
};

typedef int(*Parser)(void* args, Arg arg);

int parseCommandLine(int argc, char* const argv[], const Option options[], int size, int printErrMsg, int POSIXLY_CORRECT, Parser f, void* args);

#ifdef __cplusplus
}

namespace ArgParser {

struct Opt {
    int index;
    const char* val;
};

struct IntSt {
    int optIdx = 1;
    int charIdx = 1;
    int nonOptIdx = 0;
    bool stopOptParsing = false;
};

bool parseArg(int argc, char* const argv[], const Option options[], int size, Opt& opt, int& argi, IntSt& st, bool printErrMsg, bool POSIXLY_CORRECT);
bool handleError(int code, bool printErrMsg, const char* name, const char* val);

template<typename F, typename... Args>
bool parseCommandLine(int argc, char* const argv[], const Option options[], int size, bool printErrMsg, bool POSIXLY_CORRECT, F&& f, Args&&... args) {
    Opt opt;
    int argi;
    IntSt st;
    bool stopParsing = false;
    bool ret = true;

    while (parseArg(argc, argv, options, size, opt, argi, st, printErrMsg, POSIXLY_CORRECT)) {
        int id = options[opt.index].id;
        const char* name = options[opt.index].name;
        const char* val = opt.val;

        auto res = std::forward<F>(f)(std::forward<Args>(args)..., Arg{ id, val });
        ret = handleError(res, printErrMsg, name, val);

        if (res != ESuccess) {
            stopParsing = true;
            break;
        }
    }

    if (argi < argc && !stopParsing)
        ret = false;

    return ret;
}

}
#endif

#endif
